import { merge, Subject } from "rxjs";
import { Report } from "@oboku/reader";
import { takeUntil, tap } from "rxjs/operators";
const SHOULD_NOT_LAYOUT = false;
const HIGHLIGHT_ID_PREFIX = `oboku-reader-enhancer-highlights`;
let uniqueId = 0;
export const createHighlightsEnhancer = ({ highlights: initialHighlights }) => (next) => (options) => {
    const reader = next(options);
    const highlights$ = new Subject();
    let highlights = [];
    const getRangeForHighlight = (overlayElement, anchor, focus) => {
        const range = overlayElement.ownerDocument.createRange();
        try {
            range.setStart(anchor.node, anchor.offset || 0);
            range.setEnd(focus.node, focus.offset || 0);
        }
        catch (e) {
            Report.error(e);
        }
        return range;
    };
    const drawHighlight = (overlayElement, highlight) => {
        var _a;
        const { node: anchorNode, offset: anchorOffset } = reader.resolveCfi(highlight.anchorCfi) || {};
        const { node: focusNode, offset: focusOffset } = reader.resolveCfi(highlight.focusCfi) || {};
        if (anchorNode && focusNode) {
            (_a = highlight.element) === null || _a === void 0 ? void 0 : _a.remove();
            const range = getRangeForHighlight(overlayElement, { node: anchorNode, offset: anchorOffset }, { node: focusNode, offset: focusOffset });
            highlight.text = range.toString();
            const rectElements = Array.from(range.getClientRects()).map((domRect) => {
                const rectElt = overlayElement.ownerDocument.createElement('div');
                rectElt.style.cssText = `
            position: absolute;
            width: ${domRect.width}px;
            height: ${domRect.height}px;
            top: ${domRect.top}px;
            left: ${domRect.left}px;
            background-color: green;
            opacity: 50%;
          `;
                rectElt.setAttribute(`data-${HIGHLIGHT_ID_PREFIX}`, ``);
                return rectElt;
            });
            const containerElement = overlayElement.ownerDocument.createElement('div');
            containerElement.style.cssText = `
          pointer-events: auto;
        `;
            highlight.element = containerElement;
            rectElements.forEach(el => containerElement.appendChild(el));
            overlayElement.appendChild(containerElement);
            containerElement.addEventListener(`click`, () => {
                highlights$.next({ type: `onHighlightClick`, data: highlight });
            });
        }
    };
    const drawHighlightsForItem = (overlayElement, itemIndex) => {
        highlights.forEach((highlight) => {
            if (highlight.readingItemIndex === itemIndex) {
                drawHighlight(overlayElement, highlight);
            }
        });
    };
    const _add = (highlight) => {
        const cfiMetaInfo = reader.getCfiMetaInformation(highlight.anchorCfi);
        const newHighlight = Object.assign(Object.assign({}, highlight), { readingItemIndex: cfiMetaInfo === null || cfiMetaInfo === void 0 ? void 0 : cfiMetaInfo.readingItemIndex, id: uniqueId++ });
        highlights.push(newHighlight);
        if (newHighlight.readingItemIndex !== undefined) {
            reader.manipulateReadingItems(({ index, overlayElement }) => {
                if (index !== newHighlight.readingItemIndex)
                    return SHOULD_NOT_LAYOUT;
                drawHighlight(overlayElement, newHighlight);
                return SHOULD_NOT_LAYOUT;
            });
        }
        return highlight;
    };
    const add = (highlight) => {
        _add(highlight);
        highlights$.next({ type: `onUpdate`, data: highlights });
    };
    const remove = (id) => {
        highlights = highlights.filter(highlight => {
            var _a;
            if (highlight.id === id) {
                (_a = highlight.element) === null || _a === void 0 ? void 0 : _a.remove();
            }
            return highlight.id !== id;
        });
        highlights$.next({ type: `onUpdate`, data: highlights });
    };
    const isHighlightClicked = (event) => {
        if (event.target instanceof HTMLElement) {
            return event.target.hasAttribute(`data-${HIGHLIGHT_ID_PREFIX}`);
        }
        return false;
    };
    const initialHighlights$ = reader.$.ready$
        .pipe(tap(() => {
        initialHighlights.forEach(_add);
        if (initialHighlights.length > 0) {
            highlights$.next({ type: `onUpdate`, data: highlights });
        }
    }));
    const refreshHighlights$ = reader.$.layout$
        .pipe(tap(() => {
        reader.manipulateReadingItems(({ overlayElement, index }) => {
            drawHighlightsForItem(overlayElement, index);
            return SHOULD_NOT_LAYOUT;
        });
    }));
    merge(initialHighlights$, refreshHighlights$)
        .pipe(takeUntil(reader.$.destroy$))
        .subscribe();
    return Object.assign(Object.assign({}, reader), { highlights: {
            add,
            remove,
            isHighlightClicked,
            $: highlights$.asObservable(),
        } });
};
